# -*- coding: utf-8 -*-
"""
Created on Wed Mar  8 10:02:01 2023

@author: Jung Min Ahn
"""
import numpy as np
# import xgboost as xgb
# import lightgbm as lgbm
# import catboost as ctb
import pandas as pd
from xgboost import XGBRegressor
from lightgbm import LGBMRegressor
from catboost import CatBoostRegressor
import matplotlib.pyplot as plt
from sklearn.metrics import r2_score, mean_squared_error
import torch
from torch.autograd import Variable 
# from sklearn.metrics import mean_absolute_error



best_ite = 50 # iteration
early_stop = 300
datanum = 440 # total data number
SEED = 50

###---------------------------------

file_nm = "./Input.xlsx"
sheet_nm = "CH"
df = pd.read_excel(file_nm,sheet_nm, engine='openpyxl')
X = df.drop(columns=['order','date'])
y = df.iloc[:, 2:3]
input_size = len(X.columns)

from sklearn.preprocessing import StandardScaler, MinMaxScaler
MIN = X.min()
MAX = X.max()
ss = StandardScaler()
mm = MinMaxScaler()

X_ss = ss.fit_transform(X)
y_mm = mm.fit_transform(y)

##----------------------Boost----------------------
# Train Data
x_train = X_ss[:datanum, :].astype(float)
y_train = y_mm[1:datanum+1].astype(float)

# Test Data
x_test = X_ss[datanum:, :].astype(float)
y_test = y_mm[datanum:].astype(float)
OBS = y_test[1:].astype(float)


def RMSE(y_true, y_pred):
    return np.sqrt(mean_squared_error(y_true, y_pred))

# Define SMAPE loss function
def SMAPE(y_test, y_pred):
    y_test, y_pred = np.array(y_test), np.array(y_pred)
    return np.mean((np.abs(y_test-y_pred))/(np.abs(y_test) + np.abs(y_pred))) * 100

def MAPE(y_test, y_pred):
      y_test, y_pred = np.array(y_test), np.array(y_pred)
      return np.mean(np.abs((y_test - y_pred) / y_test)) * 100

# custom objective function for forcing model not to underestimate
def weighted_mse(alpha = 1):
    def weighted_mse_fixed(label, pred):
        residual = (label - pred).astype("float")
        grad = np.where(residual>0, -2*alpha*residual, -2*residual)
        hess = np.where(residual>0, 2*alpha, 2.0)
        return grad, hess
    return weighted_mse_fixed

#import hyperopt
from hyperopt import fmin, tpe, hp, STATUS_OK, Trials

# regularization candiate define
reg_candidate = [1e-5, 1e-4, 1e-3, 1e-2, 0.1, 1, 5, 10, 100]

# Choose hyperparameter domain to search over
xgb_space = {
        'max_depth':hp.choice('max_depth', np.arange(10, 25, 1, dtype=int)),
        'learning_rate': hp.quniform ('learning_rate', 0.01, 0.05, 0.005),      
        'reg_alpha' : hp.choice('reg_alpha', reg_candidate),
        'reg_lambda' : hp.choice('reg_lambda', reg_candidate),        
        'n_estimators':hp.choice('n_estimators', np.arange(1000, 10000, 10, dtype=int)),
        'colsample_bytree':hp.quniform('colsample_bytree', 0.5, 1.0, 0.1),
        'min_child_weight':hp.choice('min_child_weight', np.arange(250, 350, 10, dtype=int)),
        'subsample':hp.quniform('subsample', 0.7, 0.9, 0.1),
        'eta':hp.quniform('eta', 0.1, 0.3, 0.1),
        
        'objective':'reg:squarederror',
        
        'tree_method':'gpu_hist',
        #'eval_metric': 'rmse',
    }

lgbm_space = {
    'max_depth': hp.choice('max_depth_lgbm', np.arange(10, 25, 1, dtype=int)),
    'learning_rate': hp.quniform('learning_rate_lgbm', 0.01, 0.05, 0.005),
    'reg_alpha': hp.choice('reg_alpha_lgbm', reg_candidate),
    'reg_lambda': hp.choice('reg_lambda_lgbm', reg_candidate),
    'n_estimators': hp.choice('n_estimators_lgbm', np.arange(1000, 10000, 10, dtype=int)),
    'colsample_bytree': hp.quniform('colsample_bytree_lgbm', 0.5, 1.0, 0.1),
    'min_child_weight': hp.choice('min_child_weight_lgbm', np.arange(250, 350, 10, dtype=int)),
    'subsample': hp.quniform('subsample_lgbm', 0.7, 0.9, 0.1),
    'eta': hp.quniform('eta_lgbm', 0.1, 0.3, 0.1),
    'objective': 'regression',
    'boosting_type': 'gbdt',
    'metric': 'rmse',
    'device_type': 'gpu'
}

cat_space={'max_depth': hp.quniform("max_depth", 3, 9, 1),
       #'depth': hp.choice('depth_cat', np.arange(10, 25, 1)),
       #'iterations': hp.choice('iterations_cat', np.arange(1000, 10000, 10)),
       'learning_rate': hp.quniform ('learning_rate', 0.01, 0.05, 0.005),
       'l2_leaf_reg': hp.quniform ('l2_leaf_reg', 3,8,1),
       'border_count': hp.quniform ('border_count', 32,255, 10),
       'n_estimators': hp.quniform('n_estimators', 1000, 10000, 10),
       'colsample_bylevel': hp.quniform('colsample_bylevel', 0.3, 0.8, 0.1),
       'bagging_temperature': hp.uniform('bagging_temperature', 0.0, 10.0),
       'subsample': hp.uniform('subsample', 0.5, 1.0),
       'min_data_in_leaf': hp.choice('min_data_in_leaf', [1, 5, 10, 20, 30]),
       'verbose': False,
       'eval_metric': 'RMSE',
       'loss_function': 'RMSE',
       'task_type': 'GPU'
}

# Boost function define
def XGB_hyperparameter_tuning(xgb_space):
    model1=XGBRegressor(n_estimators =xgb_space['n_estimators'], 
                       eta = xgb_space['eta'],                            
                       max_depth = xgb_space['max_depth'], 
                       learning_rate = xgb_space['learning_rate'],
                       reg_alpha = xgb_space['reg_alpha'],
                       reg_lambda = xgb_space['reg_lambda'],
                       subsample = xgb_space['subsample'],
                       colsample_bytree = xgb_space['colsample_bytree'], 
                       min_child_weight = xgb_space['min_child_weight'],
                       random_state=SEED,                     
                      )
    
    evaluation = [(x_train, y_train), (x_test, y_test)]

    model1.fit(x_train, y_train,
              eval_set=evaluation, 
              #eval_metric="rmse",
              eval_metric="rmse",
              early_stopping_rounds=early_stop,
              verbose=0)

    pred = model1.predict(x_test)
    rmse= RMSE(y_test, pred)    

    return {'loss':rmse, 'status': STATUS_OK, 'model': model1}

def LGBM_hyperparameter_tuning(lgbm_space):
    model2=LGBMRegressor(n_estimators =lgbm_space['n_estimators'], 
                       eta = lgbm_space['eta'],                          
                       max_depth = lgbm_space['max_depth'], 
                       learning_rate = lgbm_space['learning_rate'],
                       reg_alpha = lgbm_space['reg_alpha'],
                       reg_lambda = lgbm_space['reg_lambda'],
                       subsample = lgbm_space['subsample'],
                       colsample_bytree = lgbm_space['colsample_bytree'], 
                       min_child_weight = lgbm_space['min_child_weight'],
                       random_state=SEED, 
                      )
    
    evaluation = [(x_train, y_train), (x_test, y_test)]
    
    model2.fit(x_train, y_train,
              eval_set=evaluation, 
              #eval_metric="rmse",
              eval_metric="rmse",
              early_stopping_rounds=early_stop,
              verbose=0)

    pred = model2.predict(x_test)
    rmse= RMSE(y_test, pred)    
    
    return {'loss':rmse, 'status': STATUS_OK, 'model': model2}

def CAT_hyperparameter_tuning(cat_space):
    model3=CatBoostRegressor(n_estimators =cat_space['n_estimators'], 
                       max_depth = cat_space['max_depth'],
                       #depth = cat_space['depth'], 
                       #iterations = cat_space['iterations'], 
                       learning_rate = cat_space['learning_rate'],
                       l2_leaf_reg = cat_space['l2_leaf_reg'],
                       border_count = cat_space['border_count'],
                       colsample_bylevel = cat_space['colsample_bylevel'], 
                       min_data_in_leaf = cat_space['min_data_in_leaf'],
                       bagging_temperature = cat_space['bagging_temperature'],
                       subsample = cat_space['subsample'],
                       eval_metric="RMSE",
                       random_state=SEED 
                      )
    
    evaluation = [(x_train, y_train), (x_test, y_test)]
   
    model3.fit(x_train, y_train,
              eval_set=evaluation,
              early_stopping_rounds=early_stop,
              verbose=0)

    pred = model3.predict(x_test)
    rmse= RMSE(y_test, pred)    
    
    return {'loss':rmse, 'status': STATUS_OK, 'model': model3}

# XGBoost Trials declare
trials = Trials()
# best parameter return
xgb_best = fmin(fn=XGB_hyperparameter_tuning,
            space=xgb_space,
            algo=tpe.suggest,
            max_evals=best_ite,
            trials=trials)

# LightGBM Trials declare
trials = Trials()
# best parameter return
LGBM_best = fmin(fn=LGBM_hyperparameter_tuning,
            space=lgbm_space,
            algo=tpe.suggest,
            max_evals=best_ite,
            trials=trials)

# CatBoost Trials declare
trials = Trials()
# best parameter return
CAT_best = fmin(fn=CAT_hyperparameter_tuning,
            space=cat_space,
            algo=tpe.suggest,
            max_evals=best_ite,
            trials=trials)

####------------------------XGBRegressor------------------------------------

xgb_preds = np.array([]) 
xgb_pred_df = pd.DataFrame() 
    
for seed in [0,1,2,3,4,5]: 
        
        xgb_model = XGBRegressor(**xgb_best, tree_method = 'gpu_hist', predictor= 'gpu_predictor')
        xgb_model.fit(x_train, y_train) 

        XGB_y_pred = xgb_model.predict(x_test)
        xgb_pred_df.loc[:,seed] = XGB_y_pred   
        
xgb_pred = xgb_pred_df.mean(axis=1)        
xgb_preds = np.append(xgb_preds, xgb_pred)   
print(RMSE(y_test, xgb_preds))

####-------------------------LGBMRegressor--------------------------------------
lgbm_preds = np.array([]) 
lgbm_pred_df = pd.DataFrame()   
    
for seed in [0,1,2,3,4,5]: 

        lgbm_model = LGBMRegressor(**LGBM_best, tree_method = 'gpu_hist', predictor= 'gpu_predictor')
        lgbm_model.fit(x_train, y_train)

        lgbm_y_pred = lgbm_model.predict(x_test)
        lgbm_pred_df.loc[:,seed] = lgbm_y_pred   

lgbm_pred = lgbm_pred_df.mean(axis=1)        
lgbm_preds = np.append(lgbm_preds, lgbm_pred)   
print(RMSE(y_test, lgbm_preds))


####-------------------------CatBoostRegressor--------------------------------------
cat_preds = np.array([]) 
cat_pred_df = pd.DataFrame()   
    
for seed in [0,1,2,3,4,5]: 

        ctb_model = CatBoostRegressor(**CAT_best)
        ctb_model.fit(x_train, y_train)

        cat_y_pred = ctb_model.predict(x_test)
        cat_pred_df.loc[:,seed] = cat_y_pred   
        
cat_pred = cat_pred_df.mean(axis=1)        
cat_preds = np.append(cat_preds, cat_pred)       
print(RMSE(y_test, cat_preds))

from sklearn.model_selection import KFold
from sklearn.linear_model import LinearRegression

####------------------------- KFold ensemble-----------------------------------------------

def stack_models(models, meta_model, X_train, y_train, X_test, y_test, n_folds=5):
    # Initialize arrays to hold base model predictions
    train_preds = np.zeros((X_train.shape[0], len(models)))
    test_preds = np.zeros((X_test.shape[0], len(models)))
    
    # Train base models using KFold
    kf = KFold(n_splits=n_folds, shuffle=True, random_state=42)
    for i, model in enumerate(models):
        for train_index, val_index in kf.split(X_train, y_train):
            model.fit(X_train[train_index], y_train[train_index])
            train_preds[val_index, i] = model.predict(X_train[val_index])
        # Predict test data for the current model
        test_preds[:, i] = model.predict(X_test)
    
    # Train meta model on base model predictions
    meta_model.fit(train_preds, y_train)
    
    # Generate final predictions using meta model and base model predictions
    final_train_preds = meta_model.predict(train_preds)
    final_test_preds = meta_model.predict(test_preds)
    
    return final_train_preds, final_test_preds

base_models = [xgb_model, lgbm_model, ctb_model]
meta_model = LinearRegression()

final_train_preds, final_test_preds = stack_models(base_models, meta_model,
x_train, y_train, x_test, y_test)


####-------------------Data Reverse transformation--------------------------------------
xgb_preds2 = xgb_preds.reshape(-1,1)
lgbm_preds2 = lgbm_preds.reshape(-1,1)
cat_preds2 = cat_preds.reshape(-1,1)
stacked_pred2 = final_test_preds.reshape(-1,1)
XGbooster_predict = mm.inverse_transform(xgb_preds2) #reverse transformation
Lgbmbooster_predict = mm.inverse_transform(lgbm_preds2) #reverse transformation
Catbooster_predict = mm.inverse_transform(cat_preds2) #reverse transformation
stacked_predict = mm.inverse_transform(stacked_pred2) #reverse transformation
dataY_plot = mm.inverse_transform(y_test)

####------------------- Graph-------------------------------------------------
plt.figure(figsize=(15,6)) #plotting
plt.plot(dataY_plot, label='Actuall Data') #actual plot
plt.plot(XGbooster_predict, label='Predicted Data') #predicted plot
plt.title('XGBoost Time-Series Prediction')
plt.legend()
plt.show() 

plt.figure(figsize=(15,6)) #plotting
plt.plot(dataY_plot, label='Actuall Data') #actual plot
plt.plot(Lgbmbooster_predict, label='Predicted Data') #predicted plot
plt.title('LGBMboost Time-Series Prediction')
plt.legend()
plt.show() 

plt.figure(figsize=(15,6)) #plotting
plt.plot(dataY_plot, label='Actuall Data') #actual plot
plt.plot(Catbooster_predict, label='Predicted Data') #predicted plot
plt.title('CATboost Time-Series Prediction')
plt.legend()
plt.show() 

plt.figure(figsize=(15,6)) #plotting
plt.plot(dataY_plot, label='Actuall Data') #actual plot
plt.plot(stacked_predict, label='Predicted Data') #predicted plot
plt.title('Stacking Ensemble Time-Series Prediction')
plt.legend()
plt.show() 

# Ensemble value calculate
BOOSTAVG = (XGbooster_predict + Lgbmbooster_predict + Catbooster_predict) / 3

plt.figure(figsize=(15,6)) #plotting
plt.plot(dataY_plot, label='Actuall Data') #actual plot
plt.plot(BOOSTAVG, label='Predicted Data') #predicted plot
plt.title('Bagging Ensemble Time-Series Prediction')
plt.legend()
plt.show()

### result
OBS = dataY_plot
XGB = XGbooster_predict
LGBM = Lgbmbooster_predict
CAT = Catbooster_predict
Stackings = stacked_predict
ENSEMBLE = BOOSTAVG

# R2, MAE, RMSE
XGB_R2 = np.round(r2_score(OBS, XGB),2)
XGB_MAE = np.round(mean_squared_error(OBS, XGB),1)
XGB_RMSE = np.round(np.sqrt(mean_squared_error(OBS, XGB)),1)
print("XGB : R2 {} || MAE : {} || RMSE : {}".format(XGB_R2, XGB_MAE, XGB_RMSE))
LGBM_R2 = np.round(r2_score(OBS, LGBM),2)
LGBM_MAE = np.round(mean_squared_error(OBS, LGBM),1)
LGBM_RMSE = np.round(np.sqrt(mean_squared_error(OBS, LGBM)),1)
print("LGBM : R2 {} || MAE : {} || RMSE : {}".format(LGBM_R2, LGBM_MAE, LGBM_RMSE))
CAT_R2 = np.round(r2_score(OBS, CAT),2)
CAT_MAE = np.round(mean_squared_error(OBS, CAT),1)
CAT_RMSE = np.round(np.sqrt(mean_squared_error(OBS, CAT)),1)
print("CAT : R2 {} || MAE : {} || RMSE : {}".format(CAT_R2, CAT_MAE, CAT_RMSE))
Stackings_R2 = np.round(r2_score(OBS, Stackings),2)
Stackings_MAE = np.round(mean_squared_error(OBS, Stackings),1)
Stackings_RMSE = np.round(np.sqrt(mean_squared_error(OBS, Stackings)),1)
print("Stacking : R2 {} || MAE : {} || RMSE : {}".format(Stackings_R2, Stackings_MAE, Stackings_RMSE))
BOOSTAVG_R2 = np.round(r2_score(OBS, BOOSTAVG),2)
BOOSTAVG_MAE = np.round(mean_squared_error(OBS, BOOSTAVG),1)
BOOSTAVG_RMSE = np.round(np.sqrt(mean_squared_error(OBS, BOOSTAVG)),1)
print("Bagging : R2 {} || MAE : {} || RMSE : {}".format(BOOSTAVG_R2, BOOSTAVG_MAE, BOOSTAVG_RMSE))
ENSEMBLE_R2 = np.round(r2_score(OBS, ENSEMBLE),2)
ENSEMBLE_MAE = np.round(mean_squared_error(OBS, ENSEMBLE),1)
ENSEMBLE_RMSE = np.round(np.sqrt(mean_squared_error(OBS, ENSEMBLE)),1)
print("ENSEMBLE : R2 {} || MAE : {} || RMSE : {}".format(ENSEMBLE_R2, ENSEMBLE_MAE, ENSEMBLE_RMSE))






